#!/usr/bin/python3
"""
// © University of Southampton IT Innovation Centre, 2018
//
// Copyright in this software belongs to University of Southampton
// IT Innovation Centre of Gamma House, Enterprise Road,
// Chilworth Science Park, Southampton, SO16 7NS, UK.
//
// This software may not be used, sold, licensed, transferred, copied
// or reproduced in whole or in part in any manner or form or in or
// on any media by any person other than in accordance with the terms
// of the Licence Agreement supplied with the software, or otherwise
// without the prior written consent of the copyright owners.
//
// This software is distributed WITHOUT ANY WARRANTY, without even the
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE, except where stated in the Licence Agreement supplied with
// the software.
//
//      Created By :            Nikolay Stanchev
//      Created Date :          15-05-2018
//      Created for Project :   FLAME
"""

from pyramid import testing
from pyramid.httpexceptions import HTTPBadRequest
from time import sleep
from clmcservice.utilities import CONF_FILE_ATTRIBUTE, CONF_OBJECT, AGGREGATOR_CONFIG_SECTION, CONFIG_ATTRIBUTES, PROCESS_ATTRIBUTE, RUNNING_FLAG, MALFORMED_FLAG, URL_REGEX
import pytest
import os
import signal
import configparser


class TestAggregatorAPI(object):
    """
    A pytest-implementation test for the aggregator API calls
    """

    @pytest.fixture(autouse=True)
    def app_config(self):
        """
        A fixture to implement setUp/tearDown functionality for all tests by initializing configuration structure for the web service
        """

        self.registry = testing.setUp()
        config = configparser.ConfigParser()
        config[AGGREGATOR_CONFIG_SECTION] = {'aggregator_report_period': 5, 'aggregator_database_name': 'CLMCMetrics', 'aggregator_database_url': "http://172.40.231.51:8086"}
        self.registry.add_settings({'configuration_object': config, 'aggregator_running': False, 'malformed': False, 'configuration_file_path': "/etc/flame/clmc/service.conf"})

        yield

        testing.tearDown()

    def test_GET_config(self):
        """
        Tests the GET method for the configuration of the aggregator.
        """

        from clmcservice.views import AggregatorConfig  # nested import so that importing the class view is part of the test itself

        assert int(self.registry.get_settings()[CONF_OBJECT][AGGREGATOR_CONFIG_SECTION].get('aggregator_report_period')) == 5, "Initial report period is 5 seconds."
        assert self.registry.get_settings()[CONF_OBJECT][AGGREGATOR_CONFIG_SECTION].get('aggregator_database_name') == 'CLMCMetrics', "Initial database name the aggregator uses is CLMCMetrics."
        assert self.registry.get_settings()[CONF_OBJECT][AGGREGATOR_CONFIG_SECTION].get('aggregator_database_url') == "http://172.40.231.51:8086", "Initial aggregator url is http://172.40.231.51:8086"

        request = testing.DummyRequest()
        response = AggregatorConfig(request).get()

        assert response == {'aggregator_report_period': 5,
                            'aggregator_database_name': 'CLMCMetrics',
                            'aggregator_database_url': "http://172.40.231.51:8086"}, "Response must be a dictionary representing a JSON object with the correct configuration data of the aggregator."

        assert int(self.registry.get_settings()[CONF_OBJECT][AGGREGATOR_CONFIG_SECTION].get('aggregator_report_period')) == 5, "A GET request must not modify the aggregator configuration data."
        assert self.registry.get_settings()[CONF_OBJECT][AGGREGATOR_CONFIG_SECTION].get('aggregator_database_name') == 'CLMCMetrics', "A GET request must not modify the aggregator configuration data."
        assert self.registry.get_settings()[CONF_OBJECT][AGGREGATOR_CONFIG_SECTION].get('aggregator_database_url') == "http://172.40.231.51:8086", "A GET request must not modify the aggregator configuration data."

    @pytest.mark.parametrize("input_body, output_value", [
        ('{"aggregator_report_period": 10, "aggregator_database_name": "CLMCMetrics", "aggregator_database_url": "http://171.40.231.51:8086"}',
         {'aggregator_report_period': 10, 'aggregator_database_name': "CLMCMetrics", 'aggregator_database_url': "http://171.40.231.51:8086"}),
        ('{"aggregator_report_period": 15, "aggregator_database_name": "CLMCMetrics", "aggregator_database_url": "http://172.50.231.51:8086"}',
         {'aggregator_report_period': 15, 'aggregator_database_name': "CLMCMetrics", 'aggregator_database_url': "http://172.50.231.51:8086"}),
        ('{"aggregator_report_period": 20, "aggregator_database_name": "CLMCMetrics", "aggregator_database_url": "http://172.60.231.51:8086"}',
         {'aggregator_report_period': 20, 'aggregator_database_name': "CLMCMetrics", 'aggregator_database_url': "http://172.60.231.51:8086"}),
        ('{"aggregator_report_period": 25, "aggregator_database_name": "CLMCMetrics", "aggregator_database_url": "http://172.60.231.51:8086"}',
         {'aggregator_report_period': 25, 'aggregator_database_name': "CLMCMetrics", 'aggregator_database_url': "http://172.60.231.51:8086"}),
        ('{"aggregator_report_period": 200, "aggregator_database_name": "CLMCMetrics", "aggregator_database_url": "https://172.50.231.51:8086"}',
         {'aggregator_report_period': 200, 'aggregator_database_name': "CLMCMetrics", 'aggregator_database_url': "https://172.50.231.51:8086"}),
        ('{"aggregator_report_period": 150, "aggregator_database_name": "CLMCMetrics", "aggregator_database_url": "https://localhost:8086"}',
         {'aggregator_report_period': 150, 'aggregator_database_name': "CLMCMetrics", 'aggregator_database_url': "https://localhost:8086"}),
        ("{aggregator_report_period: 2hb5, aggregator_database_name: CLMCMetrics, aggregator_database_url: http://172.60.231.51:8086}", None),
        ("{aggregator_report_period: 250-, aggregator_database_name: CLMCMetrics, aggregator_database_url: http://172.60.231.52:8086}", None),
        ("{aggregator_report_period: 25, aggregator_database_name: CLMCMetrics, aggregator_database_url: ftp://172.60.231.51:8086}", None),
        ("{aggregator_report_period: 25, aggregator_database_name: CLMCMetrics, aggregator_database_url: http://172.60.231.51:8086/query param}", None),
        ("{aggregator_report_period: 250, aggregator_database_name: CLMCMetrics, aggregator_database_url: http://172.60.231.52:808686}", None),
        ("{}", None),
        ("{aggregator_running: true}", None),
    ])
    def test_PUT_config(self, input_body, output_value):
        """
        Tests the PUT method for the configuration of the aggregator
        :param input_body: the input body parameter
        :param output_value: the expected output value, None for expecting an Exception
        """

        from clmcservice.views import AggregatorConfig, AggregatorController  # nested import so that importing the class view is part of the test itself

        assert not AggregatorController.is_process_running(self.registry.get_settings().get(PROCESS_ATTRIBUTE)), "Initially aggregator is not running."
        assert int(self.registry.get_settings()[CONF_OBJECT][AGGREGATOR_CONFIG_SECTION].get('aggregator_report_period')) == 5, "Initial report period is 5 seconds."
        assert self.registry.get_settings()[CONF_OBJECT][AGGREGATOR_CONFIG_SECTION].get('aggregator_database_name') == 'CLMCMetrics', "Initial database name the aggregator uses is CLMCMetrics."
        assert self.registry.get_settings()[CONF_OBJECT][AGGREGATOR_CONFIG_SECTION].get('aggregator_database_url') == "http://172.40.231.51:8086", "Initial aggregator url is http://172.40.231.51:8086"

        request = testing.DummyRequest()
        request.body = input_body.encode(request.charset)

        if output_value is not None:
            response = AggregatorConfig(request).put()
            assert response == output_value, "Response of PUT request must include the new configuration of the aggregator"

            for attribute in CONFIG_ATTRIBUTES:
                assert self.registry.get_settings()[CONF_OBJECT][AGGREGATOR_CONFIG_SECTION][attribute] == str(output_value[attribute]), "Aggregator settings configuration is not updated."

            assert not AggregatorController.is_process_running(self.registry.get_settings().get(PROCESS_ATTRIBUTE)), "Aggregator running status should not be updated after a configuration update."

            # assert that the conf file is updated
            updated_conf = configparser.ConfigParser()
            conf_file = self.registry.get_settings().get(CONF_FILE_ATTRIBUTE)
            assert updated_conf.read(conf_file) == [conf_file]
            assert AGGREGATOR_CONFIG_SECTION in updated_conf.sections()

            for attribute in CONFIG_ATTRIBUTES:
                assert self.registry.get_settings()[CONF_OBJECT][AGGREGATOR_CONFIG_SECTION][attribute] == updated_conf[AGGREGATOR_CONFIG_SECTION][attribute], "Aggregator settings configuration is not updated."

        else:
            error_raised = False
            try:
                AggregatorConfig(request).put()
            except HTTPBadRequest:
                error_raised = True

            assert error_raised, "Error must be raised in case of an invalid argument."

    def test_start(self):
        """
        Tests starting the aggregator through an API call.
        """

        from clmcservice.views import AggregatorController  # nested import so that importing the class view is part of the test itself

        assert not AggregatorController.is_process_running(self.registry.get_settings().get(PROCESS_ATTRIBUTE)), "Initially aggregator is not running."
        assert self.registry.get_settings().get(PROCESS_ATTRIBUTE) is None, "Initially no aggregator process is running."

        request = testing.DummyRequest()
        input_body = '{"action": "start"}'
        request.body = input_body.encode(request.charset)

        response = AggregatorController(request).put()
        assert response == {RUNNING_FLAG: True}, "The aggregator should have been started."
        assert AggregatorController.is_process_running(self.registry.get_settings().get(PROCESS_ATTRIBUTE)), "The aggregator should have been started."
        assert self.registry.get_settings().get(PROCESS_ATTRIBUTE) is not None, "Aggregator process should have been initialized."

        # kill the started process after the test is over
        pid = request.registry.settings[PROCESS_ATTRIBUTE].pid
        os.kill(pid, signal.SIGTERM)

    def test_stop(self):
        """
        Tests stopping the aggregator through an API call.
        """

        from clmcservice.views import AggregatorController  # nested import so that importing the class view is part of the test itself

        assert not AggregatorController.is_process_running(self.registry.get_settings().get(PROCESS_ATTRIBUTE)), "Initially aggregator is not running."
        assert self.registry.get_settings().get(PROCESS_ATTRIBUTE) is None, "Initially no aggregator process is running."

        # send a start request to trigger the aggregator
        request = testing.DummyRequest()
        input_body = '{"action": "start"}'
        request.body = input_body.encode(request.charset)
        AggregatorController(request).put()
        assert self.registry.get_settings().get(PROCESS_ATTRIBUTE) is not None, "Aggregator process should have been initialized."
        assert AggregatorController.is_process_running(self.registry.get_settings().get(PROCESS_ATTRIBUTE)), "Aggregator process should have been initialized."

        # test stopping the aggregator process when it is running
        request = testing.DummyRequest()
        input_body = '{"action": "stop"}'
        request.body = input_body.encode(request.charset)

        response = AggregatorController(request).put()
        assert response == {RUNNING_FLAG: False}, "The aggregator should have been stopped."
        assert not AggregatorController.is_process_running(self.registry.get_settings().get(PROCESS_ATTRIBUTE)), "The aggregator should have been stopped."
        assert self.registry.get_settings().get(PROCESS_ATTRIBUTE) is None, "Aggregator process should have been terminated."

        sleep(2)  # put a 2 seconds timeout so that the aggregator process can terminate

        # test stopping the aggregator process when it is not running
        request = testing.DummyRequest()
        input_body = '{"action": "stop"}'
        request.body = input_body.encode(request.charset)

        response = AggregatorController(request).put()
        assert response == {RUNNING_FLAG: False}, "The aggregator should have been stopped."
        assert not AggregatorController.is_process_running(self.registry.get_settings().get(PROCESS_ATTRIBUTE)), "The aggregator should have been stopped."
        assert self.registry.get_settings().get(PROCESS_ATTRIBUTE) is None, "Aggregator process should have been terminated."

    def test_restart(self):
        """
        Tests restarting the aggregator through an API call.
        """

        from clmcservice.views import AggregatorController  # nested import so that importing the class view is part of the test itself

        assert not AggregatorController.is_process_running(self.registry.get_settings().get(PROCESS_ATTRIBUTE)), "Initially aggregator is not running."
        assert self.registry.get_settings().get(PROCESS_ATTRIBUTE) is None, "Initially no aggregator process is running."

        # test restarting the aggregator process when it is stopped
        request = testing.DummyRequest()
        input_body = '{"action": "restart"}'
        request.body = input_body.encode(request.charset)

        response = AggregatorController(request).put()
        assert response == {RUNNING_FLAG: True}, "The aggregator should have been restarted."
        assert AggregatorController.is_process_running(self.registry.get_settings().get(PROCESS_ATTRIBUTE)), "The aggregator should have been restarted."
        assert self.registry.get_settings().get(PROCESS_ATTRIBUTE), "The aggregator process should have been reinitialised."

        # test restarting the aggregator process when it is running
        request = testing.DummyRequest()
        input_body = '{"action": "restart"}'
        request.body = input_body.encode(request.charset)

        response = AggregatorController(request).put()
        assert response == {RUNNING_FLAG: True}, "The aggregator should have been restarted."
        assert AggregatorController.is_process_running(self.registry.get_settings().get(PROCESS_ATTRIBUTE)), "The aggregator should have been restarted."
        assert self.registry.get_settings().get(PROCESS_ATTRIBUTE), "The aggregator process should have been reinitialised."

        # kill the started process after the test is over
        pid = request.registry.settings[PROCESS_ATTRIBUTE].pid
        os.kill(pid, signal.SIGTERM)

    @pytest.mark.parametrize("input_body", [
        '{"action": "malformed"}',
        '{"action": true}',
        '{"action": false}',
        '{"action": 1}',
        '{invalid-json}',
        '{"action": "start", "unneeded_argument": false}',
        '{}'
    ])
    def test_malformed_actions(self, input_body):
        """
        Tests sending a malformed type of action to the aggregator through an API call.
        """

        from clmcservice.views import AggregatorController  # nested import so that importing the class view is part of the test itself

        assert not AggregatorController.is_process_running(self.registry.get_settings().get(PROCESS_ATTRIBUTE)), "Initially aggregator is not running."
        assert self.registry.get_settings().get(PROCESS_ATTRIBUTE) is None, "Initially no aggregator process is running."

        # test restarting the aggregator process when it is running
        request = testing.DummyRequest()
        input_body = input_body
        request.body = input_body.encode(request.charset)

        error_raised = False
        try:
            AggregatorController(request).put()
        except HTTPBadRequest:
            error_raised = True

        assert error_raised

    def test_GET_status(self):
        """
        Tests the GET method for the status of the aggregator.
        """

        from clmcservice.views import AggregatorController  # nested import so that importing the class view is part of the test itself

        assert not AggregatorController.is_process_running(self.registry.get_settings().get(PROCESS_ATTRIBUTE)), "Initially aggregator is not running."
        assert self.registry.get_settings().get(PROCESS_ATTRIBUTE) is None, "Initially no aggregator process is running."

        request = testing.DummyRequest()
        response = AggregatorController(request).get()

        assert response == {'aggregator_running': False}, "Response must be a dictionary representing a JSON object with the correct status data of the aggregator."

        assert not AggregatorController.is_process_running(self.registry.get_settings().get(PROCESS_ATTRIBUTE)), "A GET request must not modify the aggregator status flag."
        assert self.registry.get_settings().get(PROCESS_ATTRIBUTE) is None, "A GET request must not start the aggregator process."

        # test status with malformed configuration
        # start the aggregator
        request = testing.DummyRequest()
        input_body = '{"action": "start"}'
        request.body = input_body.encode(request.charset)
        AggregatorController(request).put()
        self.registry.get_settings()[MALFORMED_FLAG] = True

        request = testing.DummyRequest()
        response = AggregatorController(request).get()

        assert response == {'aggregator_running': True,
                            'malformed': True,
                            'comment': 'Aggregator is running in a malformed state - it uses an old version of the configuration. Please, restart it so that the updated configuration is used.'}, \
            "Response must be a dictionary representing a JSON object with the correct status data of the aggregator."

        assert AggregatorController.is_process_running(self.registry.get_settings().get(PROCESS_ATTRIBUTE)), "A GET request must not modify the aggregator status flag."
        assert self.registry.get_settings().get(MALFORMED_FLAG), "A GET request must not modify the aggregator malformed flag."
        assert self.registry.get_settings().get(PROCESS_ATTRIBUTE) is not None, "A GET request must not stop the aggregator process."

        # kill the started process after the test is over
        pid = request.registry.settings[PROCESS_ATTRIBUTE].pid
        os.kill(pid, signal.SIGTERM)

    def test_malformed_flag_behaviour(self):
        """
        Tests the behaviour of the malformed configuration flag of the aggregator when doing a sequence of API calls.
        """

        from clmcservice.views import AggregatorController, AggregatorConfig  # nested import so that importing the class view is part of the test itself

        assert not AggregatorController.is_process_running(self.registry.get_settings().get(PROCESS_ATTRIBUTE)), "Initially aggregator is not running."
        assert not self.registry.get_settings().get(MALFORMED_FLAG), "Initially aggregator is not in a malformed state"
        assert self.registry.get_settings().get(PROCESS_ATTRIBUTE) is None, "Initially no aggregator process is running."
        assert int(self.registry.get_settings()[CONF_OBJECT][AGGREGATOR_CONFIG_SECTION].get('aggregator_report_period')) == 5, "Initial report period is 5 seconds."
        assert self.registry.get_settings()[CONF_OBJECT][AGGREGATOR_CONFIG_SECTION].get('aggregator_database_name') == 'CLMCMetrics', "Initial database name the aggregator uses is CLMCMetrics."
        assert self.registry.get_settings()[CONF_OBJECT][AGGREGATOR_CONFIG_SECTION].get('aggregator_database_url') == "http://172.40.231.51:8086", "Initial aggregator url is http://172.40.231.51:8086"

        # start the aggregator with the default configuration
        request = testing.DummyRequest()
        input_body = '{"action": "start"}'
        request.body = input_body.encode(request.charset)

        response = AggregatorController(request).put()
        assert response == {RUNNING_FLAG: True}, "The aggregator should have been started."

        # update the configuration of the aggregator while it is running
        config_body = '{"aggregator_report_period": 15, "aggregator_database_name": "CLMCMetrics", "aggregator_database_url": "http://172.50.231.51:8086"}'
        output_body = {'aggregator_report_period': 15, 'aggregator_database_name': "CLMCMetrics", 'aggregator_database_url': "http://172.50.231.51:8086", 'malformed': True,
                       'comment': 'Aggregator is running in a malformed state - it uses an old version of the configuration. Please, restart it so that the updated configuration is used.'}
        request = testing.DummyRequest()
        request.body = config_body.encode(request.charset)
        response = AggregatorConfig(request).put()
        assert response == output_body, "Response of PUT request must include the new configuration of the aggregator"

        assert AggregatorController.is_process_running(self.registry.get_settings().get(PROCESS_ATTRIBUTE)), "The aggregator shouldn't be stopped when the configuration is updated."
        assert self.registry.get_settings().get(MALFORMED_FLAG), "The malformed flag should be set when the configuration is updated while the process is running."
        assert self.registry.get_settings().get(PROCESS_ATTRIBUTE) is not None, "The aggregator shouldn't be stopped when the configuration is updated."

        # check that the malformed flag has been updated through a GET call
        request = testing.DummyRequest()
        response = AggregatorController(request).get()
        assert response == {'aggregator_running': True,
                            'malformed': True,
                            'comment': 'Aggregator is running in a malformed state - it uses an old version of the configuration. Please, restart it so that the updated configuration is used.'}, \
            "Response must be a dictionary representing a JSON object with the correct status data of the aggregator."

        # restart the aggregator with the new configuration
        request = testing.DummyRequest()
        input_body = '{"action": "restart"}'
        request.body = input_body.encode(request.charset)
        response = AggregatorController(request).put()
        assert response == {RUNNING_FLAG: True}, "The aggregator should have been restarted."
        assert AggregatorController.is_process_running(self.registry.get_settings().get(PROCESS_ATTRIBUTE)), "The aggregator should have been restarted."
        assert not self.registry.get_settings().get(MALFORMED_FLAG), "The malformed flag should have been reset to False."
        assert self.registry.get_settings().get(PROCESS_ATTRIBUTE) is not None, "The aggregator should have been restarted."

        # update the configuration again while the aggregator is running
        config_body = '{"aggregator_report_period": 30, "aggregator_database_name": "CLMCMetrics", "aggregator_database_url": "http://172.50.231.51:8086"}'
        output_body = {'aggregator_report_period': 30, 'aggregator_database_name': "CLMCMetrics", 'aggregator_database_url': "http://172.50.231.51:8086", 'malformed': True,
                       'comment': 'Aggregator is running in a malformed state - it uses an old version of the configuration. Please, restart it so that the updated configuration is used.'}
        request = testing.DummyRequest()
        request.body = config_body.encode(request.charset)
        response = AggregatorConfig(request).put()
        assert response == output_body, "Response of PUT request must include the new configuration of the aggregator"

        assert AggregatorController.is_process_running(self.registry.get_settings().get(PROCESS_ATTRIBUTE)), "The aggregator shouldn't be stopped when the configuration is updated."
        assert self.registry.get_settings().get(MALFORMED_FLAG), "The malformed flag should be set when the configuration is updated while the process is running."
        assert self.registry.get_settings().get(PROCESS_ATTRIBUTE) is not None, "The aggregator shouldn't be stopped when the configuration is updated."

        # stop the aggregator - this should also reset the malformed status flag
        request = testing.DummyRequest()
        input_body = '{"action": "stop"}'
        request.body = input_body.encode(request.charset)
        response = AggregatorController(request).put()
        assert response == {RUNNING_FLAG: False}, "The aggregator should have been stopped."
        assert not AggregatorController.is_process_running(self.registry.get_settings().get(PROCESS_ATTRIBUTE)), "The aggregator should have been stopped."
        assert not self.registry.get_settings().get(MALFORMED_FLAG), "The malformed flag should have been reset to False."
        assert self.registry.get_settings().get(PROCESS_ATTRIBUTE) is None, "The aggregator should have been stopped."

    def test_unconfigured_state(self):
        """
        Tests the behaviour of the service when in unconfigured state.
        """

        from clmcservice.views import AggregatorConfig, AggregatorController

        self.registry.get_settings()[CONF_OBJECT] = None  # unconfigured state - conf object is None

        # when doing a GET for the configuration we expect a bad request if the service is in unconfigured state
        bad_request = False
        bad_request_msg = None
        try:
            request = testing.DummyRequest()
            AggregatorConfig(request).get()
        except HTTPBadRequest as err:
            bad_request = True
            bad_request_msg = err.message

        assert bad_request
        assert bad_request_msg == "Aggregator has not been configured, yet. Send a PUT request to /aggregator/config with a JSON body of the configuration."

        # when doing a PUT for the aggregator to start/stop/restart we expect a bad request if the service is in unconfigured state
        for action in ('start', 'stop', 'restart'):
            bad_request = False
            bad_request_msg = None
            try:
                request = testing.DummyRequest()
                request.body = ('{"action": "' + action + '"}').encode(request.charset)
                AggregatorController(request).put()
            except HTTPBadRequest as err:
                bad_request = True
                bad_request_msg = err.message

            assert bad_request
            assert bad_request_msg == "You must configure the aggregator before controlling it. Send a PUT request to /aggregator/config with a JSON body of the configuration."

        # configure the aggregator
        input_body = '{"aggregator_report_period": 10, "aggregator_database_name": "CLMCMetrics", "aggregator_database_url": "http://171.40.231.51:8086"}'
        output_body = {'aggregator_report_period': 10, 'aggregator_database_name': "CLMCMetrics", 'aggregator_database_url': "http://171.40.231.51:8086"}
        request = testing.DummyRequest()
        request.body = input_body.encode(request.charset)
        response = AggregatorConfig(request).put()
        assert response == output_body

        request = testing.DummyRequest()
        assert AggregatorConfig(request).get() == output_body


class TestRegexURL(object):
    """
    A pytest-implementation test for the regular expression the service uses to validate the database URL
    """

    @pytest.mark.parametrize("valid_url", [
        "http://localhost:8080/",
        "https://localhost:80/url/path",
        "https://192.168.20.20/?query=param",
        "http://custom.domain.com",
        "http://domain.net:8888/",
        "https://10.160.150.4:21",
        "http://localhost:12345",
        "http://domain.com:21/path",
        "http://domain.com:32?path",
        "http://domain.com:43#path"
    ])
    def test_valid_urls(self, valid_url):
        """
        Tests that the regular expression can detect valid URLs.

        :param valid_url: a string representing a valid URL
        """

        matched_object = URL_REGEX.match(valid_url)

        assert matched_object is not None, "The regular expression fails in validating a correct URL."

        assert matched_object.group() is not None, "The matched object should return the full-match string"

    @pytest.mark.parametrize("invalid_url", [
        "ftp://localhost:80/url/path",
        "tcp://192.168.20.20/?query=param",
        "http:/localhost:80/",
        "https//localhost:8080/",
        "https://domain:1234/url/path",
        "http://domain.com:808080/",
        "http://localhost:8-080/",
        "http://localhost:port80/",
        "http://domain.com:8080url/path",
        "http://domain.com:8080/?url path",
    ])
    def test_invalid_urls(self, invalid_url):
        """
        Tests that the regular expression can detect invalid URLs.

        :param invalid_url: a string representing an invalid URL
        """

        matched_object = URL_REGEX.match(invalid_url)

        assert matched_object is None, "The regular expression fails in detecting an invalid URL."
